"""
@ Author      : Troy Kelly
@ Date        : 23 Sept 2018
@ Description : Philips Dynalite Library - Unofficial interface for Philips Dynalite over RS485

@ Notes:        Requires a RS485 to IP gateway (Do not use the Dynalite one - use something cheaper)
"""

import asyncio
import logging
import json

LOG = logging.getLogger(__name__)


class DynetError(Exception):
    def __init__(self, message):
        self.message = message


class PacketError(Exception):
    def __init__(self, message):
        self.message = message


class DynetPacket(object):

    def __init__(self, msg=None):
        self.sync = None
        self.area = None
        self.data = []
        self.command = None
        self.join = None
        self.chk = None
        if msg is not None:
            self.fromMsg(msg)

    def toMsg(self, sync=28, area=0, command=0, data=[0, 0, 0], join=255):
        bytes = []
        bytes.append(sync)
        bytes.append(area)
        bytes.append(data[0])
        bytes.append(command)
        bytes.append(data[1])
        bytes.append(data[2])
        bytes.append(join)
        bytes.append(self.calcsum(bytes))
        self.fromMsg(bytes)

    def fromMsg(self, msg):
        bytes = []
        for byte in msg:
            bytes.append(int(byte))
        self._msg = bytes
        if(len(self._msg) > 8):
            self.excess = self._msg[7:]
            self._msg = self._msg[:7]
        if self.calcsum(self._msg) != self._msg[7]:
            raise PacketError("Failed checksum %s" % self._msg)
        self.sync = self._msg[0]
        self.area = self._msg[1]
        self.data = [self._msg[2], self._msg[4], self._msg[5]]
        self.command = self._msg[3]
        self.join = self._msg[6]
        self.chk = self._msg[7]
        if self.sync == 28:
            if self.command < 4 or (self.command > 9 and self.command < 14):
                if self.command > 3:
                    self.preset = self.command - 6
                else:
                    self.preset = self.command
                self.preset = (self.preset + (self.data[2] * 8)) + 1
                self.fade = (self.data[0] + (self.data[1] * 256)) * 0.02
            if self.command == 101:
                self.preset = self.data[0] + 1
                self.fade = (self.data[1] + (self.data[2] * 256)) * 0.02

    def toJson(self):
        return json.dumps(self.__dict__)

    def calcsum(self, msg):
        msg = msg[:7]
        return (-(sum(ord(c) for c in "".join(map(chr, msg))) % 256) & 0xFF)


class DynetEvent(object):

    def __init__(self, eventType=None, message=None, data={}, direction=None):
        self.eventType = eventType.upper() if eventType else None
        self.msg = message
        self.data = data
        self.direction = direction

    def toJson(self):
        return json.dumps(self.__dict__)


class DynetConnection(asyncio.Protocol):

    def __init__(self, connectionMade=None, connectionLost=None, receiveHandler=None, connectionPause=None, connectionResume=None, loop=None):
        self._transport = None
        self._paused = False
        self._loop = loop
        self.connectionMade = connectionMade
        self.connectionLost = connectionLost
        self.receiveHandler = receiveHandler
        self.connectionPause = connectionPause
        self.connectionResume = connectionResume

    def connection_made(self, transport):
        self._transport = transport
        self._paused = False
        if self.connectionMade is not None:
            if self._loop is None:
                self.connectionMade(transport)
            else:
                self._loop.create_task(self.connectionMade(transport))

    def connection_lost(self, exc=None):
        self._transport = None
        if self.connectionLost is not None:
            if self._loop is None:
                self.connectionLost(exc)
            else:
                self._loop.create_task(self.connectionLost(exc))

    def pause_writing(self):
        self._paused = True
        if self.connectionPause is not None:
            if self._loop is None:
                self.connectionPause()
            else:
                self._loop.create_task(self.connectionPause())

    def resume_writing(self):
        self._paused = False
        if self.connectionResume is not None:
            if self._loop is None:
                self.connectionResume()
            else:
                self._loop.create_task(self.connectionResume())

    def data_received(self, data):
        LOG.debug("Data Received: %s" % data)
        if self.receiveHandler is not None:
            if self._loop is None:
                self.receiveHandler(data)
            else:
                self._loop.create_task(self.receiveHandler(data))

    def eof_received(self):
        LOG.debug("EOF Received")


class DynetControl(object):

    def __init__(self, dynet, loop, areaDefinition=None):
        self._dynet = dynet
        self._loop = loop
        self._area = areaDefinition

    def areaPreset(self, area, preset, fade=2):
        self._loop.create_task(self._areaPreset(area=area,preset=preset,fade=fade))

    @asyncio.coroutine
    def _areaPreset(self, area, preset, fade):
        packet = DynetPacket()
        preset = preset - 1
        bank = int((preset)/8)
        opcode = preset-(bank*8)
        if opcode > 3:
            opcode = opcode + 6
        fadeLow = int(fade / 0.02) - (int((fade / 0.02) / 256) * 256)
        fadeHigh = int((fade / 0.02) / 256)
        packet.toMsg(sync=28, area=area, command=opcode, data=[fadeLow, fadeHigh, bank], join=255)
        self._dynet.write(packet)

    def areaOff(self, area, fade=2):
        self._loop.create_task(self._areaOff(area=area,fade=fade))

    @asyncio.coroutine
    def _areaOff(self, area, fade):
        packet = DynetPacket()
        if fade > 25.5:
            fade = 25.5
        if fade < 0:
            fade = 0
        packet.toMsg(sync=28, area=area, command=104, data=[255, 0, int(fade*10)], join=255)
        self._dynet.write(packet)



class Dynet(object):

    def __init__(self, host=None, port=None, broadcaster=None, onConnect=None, onDisconnect=None, loop=None):
        if host is None or port is None or loop is None:
            raise DynetError(
                'Must supply a host, port and loop for Dynet connection')
        self._host = host
        self._port = port
        self._loop = loop
        self.broadcast = broadcaster
        self._onConnect = onConnect
        self._onDisconnect = onDisconnect
        self._conn = lambda: DynetConnection(connectionMade=self._connection, connectionLost=self._disconnection,
                                             receiveHandler=self._receive, connectionPause=self._pause, connectionResume=self._resume, loop=self._loop)
        self._transport = None
        self._handlers = {}
        self._connection_retry_timer = 1
        self._paused = False
        self._inBuffer = []
        self._outBuffer = []
        self._timeout = 30

    def cleanup(self):
        self._connection_retry_timer = 1
        self._inBuffer = []
        self._outBuffer = []
        self._transport = None

    def connect(self, onConnect=None):
        asyncio.ensure_future(self._connect())

    async def _connect(self):
        LOG.debug("Connecting to Dynet on %s:%d" % (self._host, self._port))
        try:
            await asyncio.wait_for(self._loop.create_connection(self._conn, host=self._host, port=self._port), timeout=self._timeout)
        except (ValueError, OSError, asyncio.TimeoutError) as err:
            LOG.warning("Could not connect to Dynet (%s). Retrying in %d seconds",
                        err, self._connection_retry_timer)
            self._loop.call_later(self._connection_retry_timer, self.connect)
            self._connection_retry_timer = 2 * \
                self._connection_retry_timer if self._connection_retry_timer < 32 else 60

    @asyncio.coroutine
    def _receive(self, data):
        packet = DynetPacket(data)
        if hasattr(packet, 'preset'):
            event = DynetEvent(eventType='AREAPRESET', message=("Area %d Preset %d Fade %d seconds." % (packet.area, packet.preset, packet.fade)), data={'area':packet.area,'preset':packet.preset,'fade':packet.fade,'join':packet.join}, direction="IN")
            self.broadcast(event)
        else:
            LOG.debug("Dynet Inbound: %s" % packet.toJson())

    @asyncio.coroutine
    def _pause(self):
        LOG.debug("Pausing Dynet on %s:%d" % (self._host, self._port))
        # Need to schedule a resend here
        self._paused = True

    @asyncio.coroutine
    def _resume(self):
        LOG.debug("Resuming Dynet on %s:%d" % (self._host, self._port))
        # Need to schedule a resend here
        self._paused = False

    @asyncio.coroutine
    def _connection(self, transport=None):
        LOG.debug("Connected to Dynet on %s:%d" % (self._host, self._port))
        self.cleanup()
        if not transport is None:
            self._transport = transport
            if self._onConnect is not None:
                self._loop.create_task(self._onConnect(dynet=self,transport=transport))
        else:
            raise DynetError("Connected but not transport channel provided")

    @asyncio.coroutine
    def _disconnection(self, exc=None):
        LOG.debug("Disconnected from Dynet on %s:%d" %
                  (self._host, self._port))
        self.cleanup()
        if self._onDisconnect is not None:
            self._loop.create_task(self._onDisconnect(dynet=self))

        if exc is not None:
            LOG.error(exc)

    def write(self, packet=None):
        self._loop.create_task(self._write(packet))

    @asyncio.coroutine
    def _write(self, newPacket=None):
        if self._transport is None:
            raise DynetError("Must be connected to write/send messages")

        if newPacket is not None:
            self._outBuffer.append(newPacket)

        if self._paused:
            LOG.info("Connection paused - queuing packet")
            self._loop.call_later(1, self.updateLocations)

        for idx, packet in enumerate(self._outBuffer):
            msg = bytearray()
            msg.append(packet.sync)
            msg.append(packet.area)
            msg.append(packet.data[0])
            msg.append(packet.command)
            msg.append(packet.data[1])
            msg.append(packet.data[2])
            msg.append(packet.join)
            msg.append(packet.chk)
            try:
                self._transport.write(msg)
                LOG.debug("Dynet Sent: %s" % msg)
            except:
                self._logger.error("Unable to write data: %s" % msg)
            del self._outBuffer[idx]
            #self._loop.create_task(self._receive(msg))



"""
@ Author      : Troy Kelly
@ Date        : 23 Sept 2018
@ Description : Philips Dynalite Library - Unofficial interface for Philips Dynalite over RS485

@ Notes:        Requires a RS485 to IP gateway (Do not use the Dynalite one - use something cheaper)
"""

import asyncio
import logging
import json
from .dynet import Dynet, DynetControl


class BroadcasterError(Exception):
    def __init__(self, message):
        self.message = message


class PresetError(Exception):
    def __init__(self, message):
        self.message = message


class AreaError(Exception):
    def __init__(self, message):
        self.message = message


class Event(object):

    def __init__(self, eventType=None, message=None, data={}):
        self.eventType = eventType.upper() if eventType else None
        self.msg = message
        self.data = data

    def toJson(self):
        return json.dumps(self.__dict__)


class DynaliteConfig(object):

    def __init__(self, config=None):
        self.log_level = config['log_level'].upper(
        ) if 'log_level' in config else logging.INFO
        self.log_formatter = config[
            'log_formatter'] if 'log_formatter' in config else "[%(asctime)s] %(name)s {%(filename)s:%(lineno)d} %(levelname)s - %(message)s"
        self.host = config['host'] if 'host' in config else 'localhost'
        self.port = config['port'] if 'port' in config else 12345
        self.default = config['default'] if 'default' in config else {}
        self.area = {}
        self.preset = {}
        self.area = config['area'] if 'area' in config else {}
        self.preset = config['preset'] if 'preset' in config else {}


class Broadcaster(object):

    def __init__(self, listenerFunction=None, loop=None):
        if listenerFunction is None:
            raise BroadcasterError(
                "A broadcaster bust have a listener Function")
        self._listenerFunction = listenerFunction
        self._monitoredEvents = []
        self._loop = loop

    def monitorEvent(self, eventType=None):
        if eventType is None:
            raise BroadcasterError("Must supply an event type to monitor")
        eventType = eventType.upper()
        if eventType not in self._monitoredEvents:
            self._monitoredEvents.append(eventType.upper())

    def unmonitorEvent(self, eventType=None):
        if eventType is None:
            raise BroadcasterError("Must supply an event type to un-monitor")
        eventType = eventType.upper()
        if eventType in self._monitoredEvents:
            self._monitoredEvents.remove(eventType.upper())

    def update(self, event=None, dynalite=None):
        if event is None:
            return
        if event.eventType not in self._monitoredEvents and '*' not in self._monitoredEvents:
            return
        if self._loop:
            self._loop.create_task(self._callUpdater(
                event=event, dynalite=dynalite))
        else:
            self._listenerFunction(event=event, dynalite=dynalite)

    @asyncio.coroutine
    def _callUpdater(self, event=None, dynalite=None):
        self._listenerFunction(event=event, dynalite=dynalite)


class DynalitePreset(object):

    def __init__(self, name=None, value=None, fade=2, logger=None, broadcastFunction=None, area=None, dynetControl=None):
        if not value:
            raise PresetError("A preset must have a value")
        self._logger = logger
        self.active = False
        self.name = name if name else "Preset " + str(value)
        self.value = int(value)
        self.fade = float(fade)
        self.area = area
        self.broadcastFunction = broadcastFunction
        self._control = dynetControl
        if self.broadcastFunction:
            broadcastData = {
                'area': self.area.value,
                'preset': self.value,
                'name': self.area.name + ' ' + self.name,
                'state': 'OFF'
            }
            self.broadcastFunction(
                Event(eventType='newpreset', data=broadcastData))

    def turnOn(self, send=True):
        self.active = True
        if self.broadcastFunction:
            broadcastData = {
                'area': self.area.value,
                'preset': self.value,
                'name': self.area.name + ' ' + self.name,
                'state': 'ON'
            }
            self.broadcastFunction(
                Event(eventType='preset', data=broadcastData))
        if send and self._control:
            self._control.areaPreset(
                area=self.area.value, preset=self.value, fade=self.fade)
        for preset in self.area.preset:
            if self.value != preset:
                if self.area.preset[preset].active:
                    self.area.preset[preset].turnOff(send=False)

    def turnOff(self, send=True):
        self.active = False
        if self.broadcastFunction:
            broadcastData = {
                'area': self.area.value,
                'preset': self.value,
                'name': self.area.name + ' ' + self.name,
                'state': 'OFF'
            }
            self.broadcastFunction(
                Event(eventType='preset', data=broadcastData))
        if send and self._control:
            self._control.areaOff(area=self.area.value, fade=self.fade)


class DynaliteArea(object):

    def __init__(self, name=None, value=None, fade=2, areaPresets=None, defaultPresets=None, logger=None, broadcastFunction=None, dynetControl=None):
        if not value:
            raise PresetError("An area must have a value")
        self._logger = logger
        self.name = name if name else "Area " + str(value)
        self.value = int(value)
        self.fade = fade
        self.preset = {}
        self.broadcastFunction = broadcastFunction
        self._dynetControl = dynetControl
        if areaPresets:
            for presetValue in areaPresets:
                preset = areaPresets[presetValue]
                presetName = preset['name'] if 'name' in preset else None
                presetFade = preset['fade'] if 'fade' in preset else fade
                self.preset[int(presetValue)] = DynalitePreset(
                    name=presetName, value=presetValue, fade=presetFade, logger=self._logger, broadcastFunction=self.broadcastFunction, area=self, dynetControl=self._dynetControl)
        if defaultPresets:
            for presetValue in defaultPresets:
                if int(presetValue) not in self.preset:
                    preset = defaultPresets[presetValue]
                    presetName = preset['name'] if preset['name'] else None
                    presetFade = preset['fade'] if preset['fade'] else fade
                    self.preset[int(presetValue)] = DynalitePreset(
                        name=presetName, value=presetValue, fade=presetFade, logger=self._logger, broadcastFunction=self.broadcastFunction, area=self, dynetControl=self._dynetControl)

    def presetOn(self, preset, send=True):
        if preset not in self.preset:
            self.preset[preset] = DynalitePreset(
                value=preset, fade=self.fade, logger=self._logger, broadcastFunction=self.broadcastFunction, area=self, dynetControl=self._dynetControl)
        self.preset[preset].turnOn(send=send)


class Dynalite(object):

    def __init__(self, config=None, loop=None, logger=None):
        self.loop = loop if loop else asyncio.get_event_loop()
        self._logger = logger if logger else logging.getLogger(__name__)
        self._config = DynaliteConfig(config=config)
        logging.basicConfig(level=self._config.log_level,
                            format=self._config.log_formatter)

        self._configured = False

        self._listeners = []

        self.devices = {
            'area': {}
        }

        self._dynet = None
        self.control = None

    def start(self):
        self.loop.create_task(self._start())

    def connect(self):
        self.loop.create_task(self._connect())

    def processTraffic(self, event):
        self.loop.create_task(self._processTraffic(event))

    @asyncio.coroutine
    def _processTraffic(self, event):
        if event.eventType == 'AREAPRESET':
            self.devices['area'][event.data['area']
                                 ].presetOn(event.data['preset'],send=False)
        else:
            broadcastData = {
                'area': event.data['area'],
                'namename': self.devices['area'][event.data['area']].name,
                'data': event.data.toJson()
            }
            self.broadcast(
                Event(eventType='unknown', data=broadcastData))

    @asyncio.coroutine
    def _connect(self):
        self._dynet = Dynet(host=self._config.host, port=self._config.port,
                            loop=self.loop, broadcaster=self.processTraffic, onConnect=self._connected, onDisconnect=self._disconnection)
        self._dynet.connect()

    @asyncio.coroutine
    def _connected(self, dynet=None, transport=None):
        self.control = DynetControl(
            dynet, self.loop, areaDefinition=self.devices['area'])
        if not self._configured:
            self.loop.create_task(self._configure())
        self.broadcast(Event(eventType='connected', data={}))

    @asyncio.coroutine
    def _disconnection(self, dynet=None):
        self.control = None
        self.broadcast(Event(eventType='disconnected', data={}))

    def broadcast(self, event):
        self.loop.create_task(self._broadcast(event))

    @asyncio.coroutine
    def _broadcast(self, event):
        for listenerFunction in self._listeners:
            listenerFunction.update(event=event, dynalite=self)

    @asyncio.coroutine
    def _start(self):
        self.connect()

    @asyncio.coroutine
    def _configure(self):
        for areaValue in self._config.area:
            areaName = self._config.area[areaValue]['name'] if 'name' in self._config.area[areaValue] else None
            areaPresets = self._config.area[areaValue]['preset'] if 'preset' in self._config.area[areaValue] else {
            }
            areaFade = self._config.area[areaValue]['fade'] if 'fade' in self._config.area[areaValue] else None
            if areaFade is None:
                areaFade = self._config.default['fade'] if 'fade' in self._config.default else 2
            if 'nodefault' in self._config.area[areaValue] and self._config.area[areaValue]['nodefault'] == True:
                defaultPresets = None
            else:
                defaultPresets = self._config.preset

            self._logger.debug(
                "Generating Area '%d/%s' with a default fade of %d" % (int(areaValue), areaName, areaFade))
            self.devices['area'][int(areaValue)] = DynaliteArea(
                name=areaName, value=areaValue, fade=areaFade, areaPresets=areaPresets, defaultPresets=defaultPresets, logger=self._logger, broadcastFunction=self.broadcast, dynetControl=self.control)
        self._configured = True

    def addListener(self, listenerFunction=None):
        broadcaster = Broadcaster(
            listenerFunction=listenerFunction, loop=self.loop)
        self._listeners.append(broadcaster)
        return broadcaster
